/*
    semu - Sega Saturn Emulator
    Copyright (C) 1998  Chuck Mason

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Chuck Mason IV                      chuckjr@sundial.net
    477 Moffat Loop
    Oviedo, Fl 32765
*/
                                                                
#include <stdio.h>
#include <stdlib.h>
#include "memory.h"

#define OTHER /* */

unsigned char _sh2__memory[0x1000000];
unsigned char *sh2_memory = _sh2__memory;

unsigned char sh2_read_byte(unsigned long addr)
{		     
	switch((addr & 0xFF000000) >> 24) {
		case 06:
		{
			unsigned long x;
			x = addr & 0x00FFFFFF;
			addr = 0x00800000 | x;
			break;
		}
		default:
			if(addr >= 0x1000000)
				return 0;
			break;
	}
	return(*(sh2_memory+addr));
}

unsigned short sh2_read_word(unsigned long addr)
{
#ifdef OTHER
	return((sh2_read_byte(addr) << 8) | sh2_read_byte(addr + 1));
#else
	return(sh2_read_byte(addr) | (sh2_read_byte(addr + 1) << 8));
#endif
}

unsigned long sh2_read_long(unsigned long addr)
{
#ifdef OTHER
	return((sh2_read_byte(addr) << 24) |
	       (sh2_read_byte(addr + 1) << 16) | 
	       (sh2_read_byte(addr + 2) << 8) |
	       sh2_read_byte(addr + 3));
#else
	return(
		sh2_read_byte(addr) | 
		(sh2_read_byte(addr + 1) <<  8) |
		(sh2_read_byte(addr + 2) << 16) |
		(sh2_read_byte(addr + 3) << 24)
	);
#endif
}


unsigned char sh2_write_byte(unsigned long addr, unsigned long data)
{
        switch((addr & 0xFF000000) >> 24) {
                case 06:
                {
                        unsigned long x;
                        x = addr & 0x00FFFFFF;
                        addr = 0x00800000 | x;
                        break;
                }
                default:
                        if(addr >= 0x1000000)
                                return data;
                        break;
        }

	*(sh2_memory+addr) = data & 0x000000FF;
	return data;
}

unsigned short sh2_write_word(unsigned long addr, unsigned long data)
{
#ifdef OTHER
	sh2_write_byte(addr, (data & 0xFF00) >> 8);
	sh2_write_byte((addr + 1), (data & 0x00FF));
#else
	sh2_write_byte((addr + 1), (data & 0xFF00) >> 8);
	sh2_write_byte(addr, (data & 0x00FF));
#endif
	return data;
}

unsigned long sh2_write_long(unsigned long addr, unsigned long data)
{
#ifdef OTHER
	sh2_write_word(addr, (data & 0xFFFF0000) >> 16);
	sh2_write_word((addr + 2), (data & 0x0000FFFF));
#else
	sh2_write_word(addr, (data & 0x0000FFFF));
	sh2_write_word((addr + 2), (data & 0xFFFF0000) >> 16);
#endif	
	return data;
}


