/*
    semu - Sega Saturn Emulator
    Copyright (C) 1998  Chuck Mason

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Chuck Mason IV                      chuckjr@sundial.net
    477 Moffat Loop
    Oviedo, Fl 32765
*/
                                                                
#include <stdio.h>
#include "formats.h"
#include "sh2proc.h"
#include "opcodes.h"

instruction_list sh2_instruction_list[] = {
 { 0xFFFF, 0x0009, X_FORMAT, sh2_opcode_nop, "NOP" }, /* NOP */
 { 0xF00F, 0x300C, NM_FORMAT, sh2_opcode_add, "ADD R%d, R%d" }, /* ADD Rm, Rn */
 { 0xF000, 0x7000, NI_FORMAT, sh2_opcode_addi, "ADD %02X, R%d" }, /* ADD #imm, Rn */
 { 0xF00F, 0x300E, NM_FORMAT, sh2_opcode_addc, "ADDC R%d, R%d" }, /* ADDC Rm, Rn */
 { 0xF00F, 0x300F, NM_FORMAT, sh2_opcode_addv, "ADDV R%d, R%d" }, /* ADDV Rm, Rn */
 { 0xF00F, 0x2009, NM_FORMAT, sh2_opcode_and, "AND R%d, R%d" }, /* AND Rm, Rn */
 { 0xFF00, 0xC900, I_FORMAT, sh2_opcode_andi, "AND %02X, R0" }, /* AND #imm, R0 */
 { 0xFF00, 0xCD00, I_FORMAT, sh2_opcode_andm, "AND.B %02X, @(R0, GBR)" }, /* AND.B #imm, @(R0, GBR) */
 { 0xFF00, 0x8B00, D_FORMAT, sh2_opcode_bf, "BF %02X" }, /* BF TARGT_F */
 { 0xFF00, 0x8F00, D_FORMAT, sh2_opcode_bfs, "BF/S %02X" }, /* BF/S TARGT_F */
 { 0xFF00, 0x8900, D_FORMAT, sh2_opcode_bt, "BT %02X" }, /* BT TARGT_T */
 { 0xFF00, 0x8D00, D_FORMAT, sh2_opcode_bts, "BT/S %02X" }, /* BT/S TARGT_T */
 { 0xFFFF, 0x0028, X_FORMAT, sh2_opcode_clrmac, "CLRMAC" }, /* CLRMAC */

 { 0xFFFF, 0x0008, X_FORMAT, sh2_opcode_clrt, "CLRT" }, /* CLRT */
 { 0xFFFF, 0x0018, X_FORMAT, sh2_opcode_sett, "SETT" }, /* SETT */
 
 /* CMP/xx Series */
 { 0xF00F, 0x3000, NM_FORMAT, sh2_opcode_cmpeq, "CMP/EQ R%d, R%d" }, /* CMP/EQ Rm, Rn */
 { 0xF00F, 0x3003, NM_FORMAT, sh2_opcode_cmpge, "CMP/GE R%d, R%d" }, /* CMP/GE Rm, Rn */
 { 0xF00F, 0x3007, NM_FORMAT, sh2_opcode_cmpgt, "CMP/GT R%d, R%d" }, /* CMP/GT Rm, Rn */
 { 0xF00F, 0x3006, NM_FORMAT, sh2_opcode_cmphi, "CMP/HI R%d, R%d" }, /* CMP/HI Rm, Rn */
 { 0xF00F, 0x3002, NM_FORMAT, sh2_opcode_cmphs, "CMP/HS R%d, R%d" }, /* CMP/HS Rm, Rn */
 { 0xF0FF, 0x4015, N_FORMAT, sh2_opcode_cmppl,  "CMP/PL R%d" }, /* CMP/PL Rn */
 { 0xF0FF, 0x4011, N_FORMAT, sh2_opcode_cmppz,  "CMP/PZ R%d" }, /* CMP/PZ Rn */
 { 0xF00F, 0x200C, NM_FORMAT, sh2_opcode_cmpstr, "CMP/STR R%d, R%d" }, /* CMP/STR Rm, Rn */
 { 0xFF00, 0x8800, I_FORMAT, sh2_opcode_cmpim, "CMP/IM %02X, R0" }, /* CMP/IM #imm, R0 */

 /* DIV{0S,0U,1} series */
 { 0xF00F, 0x2007, NM_FORMAT, sh2_opcode_div0s, "DIV0S R%d, R%d" }, /* DIV0S Rm, Rn */
 { 0xFFFF, 0x0019, X_FORMAT, sh2_opcode_div0u, "DIV0U" }, /* DIV0U */
 { 0xF00F, 0x3004, NM_FORMAT, sh2_opcode_div1, "DIV1 R%d, R%d" }, /* DIV1 Rm, Rn */

 { 0xF00F, 0x300D, NM_FORMAT, sh2_opcode_dmuls, "DMULS.L R%d, R%d" }, /* DMULS.L Rm, Rn */ 
 { 0xF00F, 0x3005, NM_FORMAT, sh2_opcode_dmulu, "DMULU.L R%d, R%d" }, /* DMULU.L Rm, Rn */

 { 0xF0FF, 0x4010, N_FORMAT, sh2_opcode_dt, "DT R%d" }, /* DT Rn */

 { 0xF00F, 0x600E, NM_FORMAT, sh2_opcode_extsb, "EXTS.B R%d, R%d" }, /* EXTS.B Rm, Rn */
 { 0xF00F, 0x600F, NM_FORMAT, sh2_opcode_extsw, "EXTS.W R%d, R%d" }, /* EXTS.W Rm, Rn */
 { 0xF00F, 0x600C, NM_FORMAT, sh2_opcode_extub, "EXTU.B R%d, R%d" }, /* EXTU.B Rm, Rn */
 { 0xF00F, 0x600D, NM_FORMAT, sh2_opcode_extuw, "EXTU.W R%d, R%d" }, /* EXTU.W Rm, Rn */ 

 { 0xFFFF, 0x001B, X_FORMAT, sh2_opcode_sleep, "SLEEP" }, /* SLEEP */ 
 { 0xF000, 0xA000, D12_FORMAT, sh2_opcode_bra, "BRA %03X" }, /* BRA label */
 { 0xF0FF, 0x0023, M_FORMAT, sh2_opcode_braf, "BRA R%d"  }, /* BRA Rm */ 
 { 0xF000, 0xB000, D12_FORMAT, sh2_opcode_bsr, "BSR %03X" }, /* BSR label */
 { 0xF0FF, 0x0003, M_FORMAT, sh2_opcode_bsrf, "BSRF R%d"  }, /* BSRF Rm */

 { 0xF0FF, 0x402B, M_FORMAT, sh2_opcode_jmp, "JMP @R%d" }, /* JMP @Rm */
 { 0xF0FF, 0x400B, M_FORMAT, sh2_opcode_jsr, "JSR @R%d" }, /* JSR @Rm */

 /* LDC series */
 { 0xF0FF, 0x400E, M_FORMAT, sh2_opcode_ldcsr, "LDC R%d, SR" }, /* LDC Rm, SR */
 { 0xF0FF, 0x401E, M_FORMAT, sh2_opcode_ldcgbr, "LDC R%d, GBR" }, /* LDC Rm, GBR */
 { 0xF0FF, 0x402E, M_FORMAT, sh2_opcode_ldcvbr, "LDC R%d, VBR" }, /* LDC Rm, VBR */
 { 0xF0FF, 0x4007, M_FORMAT, sh2_opcode_ldcmsr, "LDC.L @R%d+, SR" }, /* LDC.L @Rm+, SR */
 { 0xF0FF, 0x4017, M_FORMAT, sh2_opcode_ldcmgbr, "LDC.L @R%d+, GBR" }, /* LDC.L @Rm+, GBR */
 { 0xF0FF, 0x4027, M_FORMAT, sh2_opcode_ldcmvbr, "LDC.L @R%d+, VBR" }, /* LDC.L @Rm+, VBR */
 /* ---------- */
 /* LDS series */ 
 { 0xF0FF, 0x400A, M_FORMAT, sh2_opcode_ldsmach, "LDS R%d, MACH" }, /* LDS Rm, MACH */
 { 0xF0FF, 0x401A, M_FORMAT, sh2_opcode_ldsmacl, "LDS R%d, MACL" }, /* LDS Rm, MACL */
 { 0xF0FF, 0x402A, M_FORMAT, sh2_opcode_ldspr, "LDS R%d, PR"   }, /* LDS Rm, PR */
 { 0xF0FF, 0x4006, M_FORMAT, sh2_opcode_ldsmmach, "LDS.L @R%d+, MACH" }, /* LDS.L @Rm+, MACH */
 { 0xF0FF, 0x4016, M_FORMAT, sh2_opcode_ldsmmacl, "LDS.L @R%d+, MACL" }, /* LDS.L @Rm+, MACL */
 { 0xF0FF, 0x4026, M_FORMAT, sh2_opcode_ldsmpr,   "LDS.L @R%d+, PR" }, /* LDS.L @Rm+, PR */
 /* ---------- */
 { 0xF00F, 0x000F, NM_FORMAT, sh2_opcode_macl, "MAC.L @R%d+, @R%d+" }, /* MAC.L @Rm+, @Rn+ */
 { 0xF00F, 0x400F, NM_FORMAT, sh2_opcode_macw, "MAC.W @R%d+, @R%d+" }, /* MAC.W @Rm+, @Rn+ */
 /* ---------- */
 /* MOV series */
 { 0xF00F, 0x6003, NM_FORMAT, sh2_opcode_mov, "MOV R%d, R%d" }, /* MOV Rm, Rn */

 { 0xF00F, 0x0004, NM_FORMAT, sh2_opcode_movbs0, "MOV.B R%d, @(R0, R%d)" }, /* MOV.B Rm, @(R0, Rn) */
 { 0xF00F, 0x0005, NM_FORMAT, sh2_opcode_movws0, "MOV.W R%d, @(R0, R%d)" }, /* MOV.W Rm, @(R0, Rn) */
 { 0xF00F, 0x0006, NM_FORMAT, sh2_opcode_movls0, "MOV.L R%d, @(R0, R%d)" }, /* MOV.L Rm, @(R0, Rn) */

 { 0xF00F, 0x2000, NM_FORMAT, sh2_opcode_movbs, "MOV.B R%d, @R%d" }, /* MOV.B Rm, @Rn */
 { 0xF00F, 0x2001, NM_FORMAT, sh2_opcode_movws, "MOV.W R%d, @R%d" }, /* MOV.W Rm, @Rn */
 { 0xF00F, 0x2002, NM_FORMAT, sh2_opcode_movls, "MOV.L R%d, @R%d" }, /* MOV.L Rm, @Rn */

 { 0xF00F, 0x6000, NM_FORMAT, sh2_opcode_movbl, "MOV.B @R%d, R%d" }, /* MOV.B @Rm, Rn */
 { 0xF00F, 0x6001, NM_FORMAT, sh2_opcode_movwl, "MOV.W @R%d, R%d" }, /* MOV.W @Rm, Rn */
 { 0xF00F, 0x6002, NM_FORMAT, sh2_opcode_movll, "MOV.L @R%d, R%d" }, /* MOV.L @Rm, Rn */

 { 0xF00F, 0x2004, NM_FORMAT, sh2_opcode_movbm, "MOV.B R%d, @-R%d" }, /* MOV.B Rm, @-Rn */
 { 0xF00F, 0x2005, NM_FORMAT, sh2_opcode_movwm, "MOV.W R%d, @-R%d" }, /* MOV.W Rm, @-Rn */
 { 0xF00F, 0x2006, NM_FORMAT, sh2_opcode_movlm, "MOV.L R%d, @-R%d" }, /* MOV.L Rm, @-Rn */

 { 0xF00F, 0x6004, NM_FORMAT, sh2_opcode_movbp, "MOV.B @R%d+, R%d" }, /* MOV.B @Rm+, Rn */
 { 0xF00F, 0x6005, NM_FORMAT, sh2_opcode_movwp, "MOV.W @R%d+, R%d" }, /* MOV.W @Rm+, Rn */
 { 0xF00F, 0x6006, NM_FORMAT, sh2_opcode_movlp, "MOV.L @R%d+, R%d" }, /* MOV.L @Rm+, Rn */
 
 { 0xF00F, 0x000C, NM_FORMAT, sh2_opcode_movbl0, "MOV.B @(R0, R%d), R%d" }, /* MOV.B @(R0, Rm), Rn */
 { 0xF00F, 0x000D, NM_FORMAT, sh2_opcode_movwl0, "MOV.W @(R0, R%d), R%d" }, /* MOV.W @(R0, Rm), Rn */
 { 0xF00F, 0x000E, NM_FORMAT, sh2_opcode_movll0, "MOV.L @(R0, R%d), R%d" }, /* MOV.L @(R0, Rm), Rn */
 
 { 0xF000, 0xE000, NI_FORMAT, sh2_opcode_movi, "MOV %02X, R%d" },  /* MOV #imm, Rn */
 { 0xF000, 0x9000, ND8_FORMAT, sh2_opcode_movwi, "MOV.W @(%02X, PC), R%d" }, /* MOV.W @(disp, PC), Rn */
 { 0xF000, 0xD000, ND8_FORMAT, sh2_opcode_movli, "MOV.L @(%02X, PC), R%d" }, /* MOV.L @(disp, PC), Rn */

 { 0xFF00, 0xC400, D_FORMAT, sh2_opcode_movblg, "MOV.B @(%02X, GBR), R0" }, /* MOV.B @(disp, GBR), R0 */
 { 0xFF00, 0xC500, D_FORMAT, sh2_opcode_movwlg, "MOV.W @(%02X, GBR), R0" }, /* MOV.W @(disp, GBR), R0 */
 { 0xFF00, 0xC600, D_FORMAT, sh2_opcode_movllg, "MOV.L @(%02X, GBR), R0" }, /* MOV.L @(disp, GBR), R0 */
 { 0xFF00, 0xC000, D_FORMAT, sh2_opcode_movbsg, "MOV.B R0, @(%02X, GBR)" }, /* MOV.B R0, @(disp, GBR) */
 { 0xFF00, 0xC100, D_FORMAT, sh2_opcode_movwsg, "MOV.B R0, @(%02X, GBR)" }, /* MOV.W R0, @(disp, GBR) */
 { 0xFF00, 0xC200, D_FORMAT, sh2_opcode_movlsg, "MOV.B R0, @(%02X, GBR)" }, /* MOV.L R0, @(disp, GBR) */
 
 { 0xFF00, 0x8000, ND4_FORMAT, sh2_opcode_movbs4, "MOV.B R0, @(%01X, R%d)" },
 { 0xFF00, 0x8100, ND4_FORMAT, sh2_opcode_movws4, "MOV.W R0, @(%01X, R%d)" },
 { 0xF000, 0x1000, NMD_FORMAT, sh2_opcode_movls4, "MOV.L R0, @(%01X, R%d)" },
 { 0xFF00, 0x8400, MD_FORMAT, sh2_opcode_movbl4, "MOV.B @(%01X, R%d), R0" },
 { 0xFF00, 0x8500, MD_FORMAT, sh2_opcode_movwl4, "MOV.W @(%01X, R%d), R0" },
 { 0xF000, 0x5000, NMD_FORMAT, sh2_opcode_movll4, "MOV.L @(%01X, R%d), R0" },

 { 0xFF00, 0xC700, D_FORMAT, sh2_opcode_mova, "MOVA @(%02X, PC), R0" },  
 { 0xF0FF, 0x0029, N_FORMAT, sh2_opcode_movt, "MOVT R%d" },
 /* NO MORE MOVs!! */
 { 0xF00F, 0x0007, NM_FORMAT, sh2_opcode_mull, "MUL.L R%d, R%d" },
 { 0xF00F, 0x200F, NM_FORMAT, sh2_opcode_muls, "MULS(.W) R%d, R%d" },
 { 0xF00F, 0x200E, NM_FORMAT, sh2_opcode_mulu, "MULU(.W) R%d, R%d" },

 { 0xF00F, 0x600B, NM_FORMAT, sh2_opcode_neg, "NEG R%d, R%d" },
 { 0xF00F, 0x600A, NM_FORMAT, sh2_opcode_negc, "NEGC R%d, R%d" },
 
 { 0xF00F, 0x6007, NM_FORMAT, sh2_opcode_not, "NOT R%d, R%d" },
 
 { 0xF00F, 0x200B, NM_FORMAT, sh2_opcode_or, "OR R%d, R%d" },
 { 0xFF00, 0xCB00, I_FORMAT, sh2_opcode_ori, "OR %02X, R0" },
 { 0xFF00, 0xCF00, I_FORMAT, sh2_opcode_orm, "OR.B %02X, @(R0, GBR)" },
 
 { 0xF0FF, 0x4024, N_FORMAT, sh2_opcode_rotcl, "ROTCL R%d" },
 { 0xF0FF, 0x4025, N_FORMAT, sh2_opcode_rotcr, "ROTCR R%d" }, 
 { 0xF0FF, 0x4004, N_FORMAT, sh2_opcode_rotl,  "ROTL R%d" },
 { 0xF0FF, 0x4005, N_FORMAT, sh2_opcode_rotr,  "ROTR R%d" },
 
 { 0xFFFF, 0x002B, X_FORMAT, sh2_opcode_rte, "RTE" },
 { 0xFFFF, 0x000B, X_FORMAT, sh2_opcode_rts, "RTS" },

 { 0xF0FF, 0x4020, N_FORMAT, sh2_opcode_shal, "SHAL R%d" },
 { 0xF0FF, 0x4021, N_FORMAT, sh2_opcode_shar, "SHAR R%d" },
 { 0xF0FF, 0x4000, N_FORMAT, sh2_opcode_shal, "SHLL R%d" },
 
 { 0xF0FF, 0x4008, N_FORMAT, sh2_opcode_shll2, "SHLL2 R%d" },
 { 0xF0FF, 0x4018, N_FORMAT, sh2_opcode_shll8, "SHLL8 R%d" },
 { 0xF0FF, 0x4028, N_FORMAT, sh2_opcode_shll16, "SHLL16 R%d" },
 
 { 0xF0FF, 0x4001, N_FORMAT, sh2_opcode_shlr, "SHLR R%d" },
 
 { 0xF0FF, 0x4009, N_FORMAT, sh2_opcode_shlr2, "SHLR2 R%d" },
 { 0xF0FF, 0x4019, N_FORMAT, sh2_opcode_shlr8, "SHLR8 R%d" },
 { 0xF0FF, 0x4029, N_FORMAT, sh2_opcode_shlr16, "SHLR16 R%d" },
 
 { 0xF0FF, 0x0002, N_FORMAT, sh2_opcode_stcsr, "STC SR, R%d" },
 { 0xF0FF, 0x0012, N_FORMAT, sh2_opcode_stcgbr, "STC GBR, R%d" },
 { 0xF0FF, 0x0022, N_FORMAT, sh2_opcode_stcvbr, "STC VBR, R%d" },
 { 0xF0FF, 0x4003, N_FORMAT, sh2_opcode_stcmsr, "STC.L SR, @-R%d" },
 { 0xF0FF, 0x4013, N_FORMAT, sh2_opcode_stcmgbr, "STC.L GBR, @-R%d" },
 { 0xF0FF, 0x4023, N_FORMAT, sh2_opcode_stcmvbr, "STC.L VBR, @-R%d" },
 
 { 0xF0FF, 0x000A, N_FORMAT, sh2_opcode_stsmach, "STS MACH, R%d" },
 { 0xF0FF, 0x001A, N_FORMAT, sh2_opcode_stsmacl, "STS MACL, R%d" },
 { 0xF0FF, 0x002A, N_FORMAT, sh2_opcode_stspr, "STS PR, R%d" },
 { 0xF0FF, 0x4002, N_FORMAT, sh2_opcode_stsmmach, "STS.L MACH, @-R%d" },
 { 0xF0FF, 0x4012, N_FORMAT, sh2_opcode_stsmmacl, "STS.L MACL, @-R%d" },
 { 0xF0FF, 0x4022, N_FORMAT, sh2_opcode_stsmpr, "STS.L PR, @-R%d" },

 { 0xF00F, 0x3008, NM_FORMAT, sh2_opcode_sub, "SUB R%d, R%d" },
 { 0xF00F, 0x300A, NM_FORMAT, sh2_opcode_subc, "SUBC R%d, R%d" },
 { 0xF00F, 0x300B, NM_FORMAT, sh2_opcode_subv, "SUBV R%d, R%d" }, 
 
 { 0xF00F, 0x6008, NM_FORMAT, sh2_opcode_swapb, "SWAP.B R%d, R%d" },
 { 0xF00F, 0x6009, NM_FORMAT, sh2_opcode_swapw, "SWAP.W R%d, R%d" },
 
 { 0xF0FF, 0x401B, N_FORMAT, sh2_opcode_tas, "TAS.B @R%d" },
 
 { 0xFF00, 0xC300, I_FORMAT, sh2_opcode_trapa, "TRAPA %02X" },
 
 { 0xF00F, 0x2008, NM_FORMAT, sh2_opcode_tst, "TST R%d, R%d" },
 { 0xFF00, 0xC800, I_FORMAT, sh2_opcode_tsti, "TST %02X, R0" },
 { 0xFF00, 0xCC00, I_FORMAT, sh2_opcode_tstm, "TST.B %02X, (R0, GBR)" },
 
 { 0xF00F, 0x200A, NM_FORMAT, sh2_opcode_xor, "XOR R%d, R%d" },
 { 0xFF00, 0xCA00, I_FORMAT, sh2_opcode_xori, "XOR %02X, R0" },
 { 0xFF00, 0xCE00, I_FORMAT, sh2_opcode_xorm, "XOR.B %02X, (R0, GBR)" },

 { 0xF00F, 0x200D, NM_FORMAT, sh2_opcode_xtrct, "XTRCT R%d, R%d" } /* yay */ 
};

static instruction current_instruction;
static unsigned long delay_instruction;
static short int delay_slot_in_progress = 0;

static void delay_slot(unsigned long npc)
{
	delay_instruction = sh2.PC;
	sh2.PC = npc;
	delay_slot_in_progress = 2;
}

void sh2_execute_instruction(unsigned long opcode)
{
	current_instruction.opcode = opcode;
	sh2_instructions[opcode].execute();

	if((delay_slot_in_progress) && (--delay_slot_in_progress) == 0) {
		sh2.PC = delay_instruction;
		delay_slot_in_progress = 0;
	}
#ifdef DEBUG
	{
		char opstr[512];
		if(sh2_invalid_opcode_occ)
			return;
			
		sh2_format_instruction_string(opcode, opstr);
		printf("%s\n", opstr);
	}
#endif
}

void sh2_format_instruction_string(unsigned long opcode, char *string)
{
	instruction sinstruct;
	unsigned long l;

	sinstruct.opcode = opcode;
	
	switch(sh2_instructions[opcode].format) {
		case X_FORMAT:
			strcpy(string, sh2_instructions[opcode].instrformat);
			break;
		case N_FORMAT:
			sprintf(string, 
				sh2_instructions[opcode].instrformat, 
				sinstruct.n.dest);
			break;
		case M_FORMAT:
			sprintf(string,
				sh2_instructions[opcode].instrformat,
				sinstruct.m.source);
			break;
		case NM_FORMAT:
			sprintf(string, 
				sh2_instructions[opcode].instrformat,
				sinstruct.nm.source,
				sinstruct.nm.dest);
			break;
		case MD_FORMAT:
			l = sinstruct.md.disp & 0x0000000F;
			sprintf(string,
				sh2_instructions[opcode].instrformat,
				l,
				sinstruct.md.source);
			break;
		case ND4_FORMAT:
			l = sinstruct.nd4.disp & 0x0000000F;
			sprintf(string,
				sh2_instructions[opcode].instrformat,
				l,
				sinstruct.nd4.dest);
			break;
		case NMD_FORMAT:
			l = sinstruct.nmd.disp & 0x0000000F;
			sprintf(string,
				sh2_instructions[opcode].instrformat,
				sinstruct.nmd.source,
				l,
				sinstruct.nmd.dest);
			break;
		case D_FORMAT:
			l = sinstruct.d.disp & 0x000000FF;
			sprintf(string,
				sh2_instructions[opcode].instrformat,
				l);
			break;
		case D12_FORMAT:
			l = sinstruct.d12.disp & 0x00000FFF;
			sprintf(string,
				sh2_instructions[opcode].instrformat,
				l);
			break;
		case ND8_FORMAT:
			l = sinstruct.nd8.disp & 0x000000FF;
			sprintf(string,
				sh2_instructions[opcode].instrformat,
				l,
				sinstruct.nd8.dest);
			break;
		case I_FORMAT:
			l = sinstruct.i.imm & 0x000000FF;
			sprintf(string,
				sh2_instructions[opcode].instrformat,
				l);
			break;
		case NI_FORMAT:
			l = sinstruct.ni.imm & 0x000000FF;
			sprintf(string,
				sh2_instructions[opcode].instrformat,
				l,
				sinstruct.ni.dest);
			break;
		default:
			l = opcode & 0x0000FFFF;
			sprintf(string, "There is an error with sh2_format_instruction_string(%04X)", l);
			break;
	}
	return;
}

/* *********************************OPCODES FOLLOW !@#%$# ********************/
void sh2_opcode_nop()
{
	sh2.PC += 2;
}

void sh2_opcode_add()
{
	sh2.R[current_instruction.nm.dest] += sh2.R[current_instruction.nm.source];
	sh2.PC += 2;
}

void sh2_opcode_addi()
{
	long i = current_instruction.ni.imm;
      
        if ((i & 0x80) == 0)
                sh2.R[current_instruction.ni.dest] += (0x000000FF & (long)i);
        else
                sh2.R[current_instruction.ni.dest] += (0xFFFFFF00 | (long)i);
	sh2.PC += 2;
}

void sh2_opcode_addc()
{
        unsigned long tmp0, tmp1;

        tmp1 = sh2.R[current_instruction.nm.dest] + sh2.R[current_instruction.nm.source];
        tmp0 = sh2.R[current_instruction.nm.dest];

        sh2.R[current_instruction.nm.dest] = tmp1 + T;

        if (tmp0 > tmp1)
                T = 1;
        else
                T = 0;

        if (tmp1 > sh2.R[current_instruction.nm.dest])
                T = 1;
        sh2.PC += 2;
}

void sh2_opcode_addv()
{
	long dest, src, ans;

	if((long)sh2.R[current_instruction.nm.dest] >= 0)
        	dest = 0;
	else
        	dest = 1;
	
	if((long)sh2.R[current_instruction.nm.source] >= 0)
        	src = 0;
	else
        	src = 1;
	
	src += dest;
	
	sh2.R[current_instruction.nm.dest] += sh2.R[current_instruction.nm.source];
	
	if((long)sh2.R[current_instruction.nm.dest] >= 0)
		ans = 0;
	else
		ans = 1;

	ans += dest;

	if (src == 0 || src == 2) {
		if (ans == 1)
                	T = 1;
		else T = 0;
	} else
		T = 0;

	sh2.PC += 2;
}

void sh2_opcode_and()
{
	sh2.R[current_instruction.nm.dest] &= sh2.R[current_instruction.nm.source];
	sh2.PC += 2;
}

void sh2_opcode_andi()
{
	sh2.R[0] &= (0x000000FF & (long)current_instruction.i.imm);
	sh2.PC += 2;
}

void sh2_opcode_andm()
{
	long temp, i = current_instruction.i.imm;
	
	temp  = (long)sh2_read_byte(sh2.GBR + sh2.R[0]);
	temp &= (0x000000FF & (long)i);

	sh2_write_byte((sh2.GBR + sh2.R[0]), temp);
	sh2.PC += 2;
}


void sh2_opcode_bf()
{
        long disp;
	long d = current_instruction.d.disp;

        if ((d & 0x80) == 0)
                disp = (0x000000FF & (long)d);
        else
                disp = (0xFFFFFF00 | (long)d);
                
        if (T == 0)
                sh2.PC = sh2.PC + (disp << 1) + 4;
        else
                sh2.PC += 2;
        
}	

void sh2_opcode_bt()
{
	long disp;
	long d = current_instruction.d.disp;

	if((d & 0x80) == 0) 
		disp = (0x000000FF & (long)d);
	else 
		disp = (0xFFFFFF00 | (long)d);

	if(T == 1) 
		sh2.PC = sh2.PC + (disp << 1) + 4;
	else 
		sh2.PC += 2;
}

void sh2_opcode_clrmac()
{
	sh2.MACH = sh2.MACL = 0;
	sh2.PC += 2;
}

void sh2_opcode_clrt()
{
	T = 0;
	sh2.PC += 2;
}

void sh2_opcode_sett()
{
	T = 1;
	sh2.PC += 2;
}

void sh2_opcode_invalid()
{
	int i;
	unsigned long op;
	
	op = (0x0000FFFF & current_instruction.opcode);

	printf("Invalid opcode (%04X)!\n", op);
	printf("R00..: %08X R01..: %08X R02..: %08X R03..: %08X\n", sh2.R[0], sh2.R[1], sh2.R[2], sh2.R[3]);
	printf("R04..: %08X R05..: %08X R06..: %08X R07..: %08X\n", sh2.R[4], sh2.R[5], sh2.R[6], sh2.R[7]);
	printf("R08..: %08X R09..: %08X R10..: %08X R11..: %08X\n", sh2.R[8], sh2.R[9], sh2.R[10], sh2.R[11]);
	printf("R12..: %08X R13..: %08X R14..: %08X R15..: %08X\n", sh2.R[10], sh2.R[11], sh2.R[12], sh2.R[13], sh2.R[14], sh2.R[15]);
	printf("MACH.: %08X MACL.: %08X GBR..: %08X VBR..: %08X\n", sh2.MACH, sh2.MACL, sh2.GBR, sh2.VBR);
	printf("PC...: %08X PR...: %08X SR...: %08X FLAG.: %c%c%c%c\n",
		sh2.PC, sh2.PR, sh2.SR, (M == 1) ? 'M' : 'm',
			    (Q == 1) ? 'Q' : 'q',
			    (S == 1) ? 'S' : 's',
			    (T == 1) ? 'T' : 't');

#ifdef DEBUG
	if(sh2_in_debugger == 0) {
		printf("Invalid instruction caught... ");
		fflush(stdout);
		sh2_debugger(0);
	}
#endif

	sh2_invalid_opcode_occ = 1;
}

void sh2_opcode_cmpeq()
{
	if(sh2.R[current_instruction.nm.dest] == sh2.R[current_instruction.nm.source]) 
		T = 1;
	else 
		T = 0;
	sh2.PC += 2;
}

void sh2_opcode_cmpge()
{
	if((long)sh2.R[current_instruction.nm.dest] >= 
           (long)sh2.R[current_instruction.nm.source])
		T = 1;
	else
		T = 0;
	sh2.PC += 2;
}
	
void sh2_opcode_cmpgt()
{
	if((long)sh2.R[current_instruction.nm.dest] > 
           (long)sh2.R[current_instruction.nm.source])
		T = 1;
	else
		T = 0;
	sh2.PC += 2;
}

void sh2_opcode_cmphi()
{
	if((unsigned long)sh2.R[current_instruction.nm.dest] >
	   (unsigned long)sh2.R[current_instruction.nm.source])
		T = 1;
	else
		T = 0;
	sh2.PC += 2;
}

void sh2_opcode_cmphs()
{
	if((unsigned long)sh2.R[current_instruction.nm.dest] >=
	   (unsigned long)sh2.R[current_instruction.nm.source])
		T = 1;
	else
		T = 0;
	sh2.PC += 2;
}

void sh2_opcode_cmppl()
{
	if((long)sh2.R[current_instruction.n.dest] > 0)
		T = 1;
	else
		T = 0;
	sh2.PC += 2;
}

void sh2_opcode_cmppz()
{
	if((long)sh2.R[current_instruction.n.dest] >= 0)
		T = 1;
	else
		T = 1;
	sh2.PC += 2;
}

void sh2_opcode_cmpstr()
{
	unsigned long temp;
	long HH, HL, LH, LL;

	temp = sh2.R[current_instruction.nm.dest] ^ sh2.R[current_instruction.nm.source];
	HH = (temp >> 12) & 0x000000FF;
	HL = (temp >>  8) & 0x000000FF;
	LH = (temp >>  4) & 0x000000FF;
	LL = temp         & 0x000000FF;

	HH = HH && HL && LH && LL;
	if(HH == 0)
		T = 1;
	else
		T = 0;
	sh2.PC += 2;
}

void sh2_opcode_cmpim()
{
	long imm;
	long i = current_instruction.i.imm;


	if((i & 0x80) == 0)
		imm = (0x000000FF & (long)i);
	else
		imm = (0xFFFFFF00 | (long)i);
	
	if(sh2.R[0] == imm) 
		T = 1;
	else
		T = 0;
	sh2.PC += 2;
}

void sh2_opcode_div0s()
{

	if((sh2.R[current_instruction.nm.dest] & 0x80000000) == 0) 
		Q = 0;
	else
		Q = 1;

	if((sh2.R[current_instruction.nm.source] & 0x80000000) == 0)
		M = 0;
	else
		M = 1;

	T = !(M == Q);
	sh2.PC += 2;
}

void sh2_opcode_div0u()
{
	Q = M = T = 0;
	sh2.PC += 2;
}

void sh2_opcode_div1()
{
	unsigned long tmp0;
	unsigned char old_q, tmp1;
	unsigned long n = current_instruction.nm.dest;
	unsigned long m = current_instruction.nm.source;

	old_q = Q;
	Q = (unsigned char)((0x80000000 & sh2.R[n]) != 0);
	
	sh2.R[n] <<= 1;
	sh2.R[n] |= (unsigned long)T;

	switch(old_q) {
		case 0:
			switch(M) {
				case 0:
					tmp0  = sh2.R[n];
					sh2.R[n] -= sh2.R[m];
					tmp1  = (sh2.R[n] > tmp0);
					switch(Q) {
						case 0:
							Q = tmp1;
							break;
						case 1:
							Q = (unsigned char)(tmp1 == 0);
							break;
					}
					break;
				case 1:
					tmp0  = sh2.R[n];
					sh2.R[n] += sh2.R[m];
					tmp1  = (sh2.R[n] < tmp0);
					switch(Q) {
						case 0:
							Q = (unsigned char)(tmp1 == 0);
							break;
						case 1:
							Q = tmp1;
							break;
					}
					break;
			}
			break;
		case 1:
			switch(M) {
				case 0:
					tmp0  = sh2.R[n];
					sh2.R[n] += sh2.R[m];
					tmp1 = (sh2.R[n] < tmp0);
					switch(Q) {
						case 0: 
							Q = tmp1;
							break;
						case 1:
							Q = (unsigned char)(tmp1 == 0);
							break;
					}
					break;
				case 1:
					tmp0  = sh2.R[n];
					sh2.R[n] -= sh2.R[m];
					tmp1 = (sh2.R[n] > tmp0);
					switch(Q) {
						case 0:	
							Q = (unsigned char)(tmp1 == 0);
							break;
						case 1:
							Q = tmp1;
							break;
					}
					break;
			}
			break;
	}
	T = (Q == M);
	sh2.PC += 2;
}

void sh2_opcode_dmuls()
{
	unsigned long n = current_instruction.nm.dest;
	unsigned long m = current_instruction.nm.source;
	
	unsigned long rnl, rnh, rml, rmh, res0, res1, res2;
	unsigned long temp0, temp1, temp2, temp3;
	long tempm, tempn, fnlml;

	tempn = (long)sh2.R[n];
	tempm = (long)sh2.R[m];

	if(tempn < 0) 
		tempn = 0 - tempn;
	if(tempm < 0)
		tempm = 0 - tempm;
	if((long)(sh2.R[n] ^ sh2.R[m]) < 0)
		fnlml = -1;
	else
		fnlml = 0;

	temp1 = (unsigned long)tempn;
	temp2 = (unsigned long)tempm;

	rnl = temp1 & 0x0000FFFF;
	rnh = (temp1 >> 16) & 0x0000FFFF;
	rml = temp2 & 0x0000FFFF;
	rmh = (temp2 >> 16) & 0x0000FFFF;

	temp0 = rml * rnl;
	temp1 = rmh * rnl;
	temp2 = rml * rnh;
	temp3 = rmh * rnh;

	res2 = 0;
	res1 = temp1 + temp2;
	
	if(res1 < temp1)
		res2 += 0x00010000;
	
	temp1 = (res1 << 16) & 0xFFFF0000;
	res0 = temp0 + temp1;
	if(res0 < temp0)
		res2++;

	res2 = res2 + ((res1 >> 16) & 0x0000FFFF) + temp3;

	if(fnlml < 0) {
		res2 = ~res2;
		if(res0 == 0)
			res2++;
		else
			res0 = (~res0) + 1;
	}
	sh2.MACH = res2;
	sh2.MACL = res0;
	sh2.PC += 2;
}

void sh2_opcode_dmulu()
{
	unsigned long m = current_instruction.nm.source;
	unsigned long n = current_instruction.nm.dest;
	unsigned long rnl, rnh, rml, rmh, res0, res1, res2;
	unsigned long temp0, temp1, temp2, temp3;

	rnl = sh2.R[n] & 0x0000FFFF;
	rnh = (sh2.R[n] >> 16) & 0x0000FFFF;
	rml = sh2.R[m] & 0x0000FFFF;
	rmh = (sh2.R[m] >> 16) & 0x0000FFFF;

	temp0 = rml * rnl;
	temp1 = rmh * rnl;
	temp2 = rml * rnh;
	temp3 = rmh * rnh;
	
	res2 = 0;
	res1 = temp1 + temp2;
	
	if(res1 < temp1)
		res2 += 0x00010000;
	
	temp1 = (res1 << 16) & 0xFFFF0000;
	res0 = temp0 + temp1;

	if(res0 < temp0) 
		res2++;

	res2 = res2 + ((res1 >> 16) & 0x0000FFFF) + temp3;

	sh2.MACH = res2;
	sh2.MACL = res0;
	sh2.PC += 2;
}

void sh2_opcode_dt()
{
	sh2.R[current_instruction.n.dest]--;
	if(sh2.R[current_instruction.n.dest] == 0)
		T = 1;
	else
		T = 0;
	sh2.PC += 2;
}

void sh2_opcode_extsb()
{
	sh2.R[current_instruction.nm.dest] = sh2.R[current_instruction.nm.source];

	if((sh2.R[current_instruction.nm.source] & 0x00000080) == 0)
		sh2.R[current_instruction.nm.dest] &= 0x000000FF;
	else
		sh2.R[current_instruction.nm.dest] |= 0xFFFFFF00;
	sh2.PC += 2;
}

void sh2_opcode_extsw()
{
	sh2.R[current_instruction.nm.dest] = sh2.R[current_instruction.nm.source];

	if((sh2.R[current_instruction.nm.source] & 0x00008000) == 0)
		sh2.R[current_instruction.nm.dest] &= 0x0000FFFF;
	else
		sh2.R[current_instruction.nm.source] |= 0xFFFF0000;
	sh2.PC += 2;
}
	
void sh2_opcode_extub()
{
	sh2.R[current_instruction.nm.dest]  = sh2.R[current_instruction.nm.source];
	sh2.R[current_instruction.nm.dest] &= 0x000000FF;
	sh2.PC += 2;
}

void sh2_opcode_extuw()
{
	sh2.R[current_instruction.nm.dest]  = sh2.R[current_instruction.nm.source];
	sh2.R[current_instruction.nm.dest] &= 0x0000FFFF;
	sh2.PC += 2;
}

void sh2_opcode_sleep()
{
	/* Basically a NOP w/o the PC increment */
	sh2_processor_sleep = 1;
	sh2.PC = sh2.PC;
}

void sh2_opcode_bfs()
{
	long disp, d = current_instruction.d.disp;
	unsigned long temp;

	temp = sh2.PC;

	if((d & 0x80) == 0) 
		disp = (0x000000FF & (long)d);
	else 
		disp = (0xFFFFFF00 | (long)d);
	if(T == 0) {
		sh2.PC = sh2.PC + (disp << 1) + 4;
		delay_slot(temp + 2);
	} else 
		sh2.PC += 2;
}

void sh2_opcode_bts()
{
	long disp;
	unsigned long temp;
	long d = current_instruction.d.disp;

	temp = sh2.PC;
	if((d & 0x80) == 0) 
		disp = (0x000000FF & (long)d);
	else 
		disp = (0xFFFFFF00 | (long)d);
	if(T == 1) {
		sh2.PC = sh2.PC + (disp << 1) + 4;
		delay_slot(temp + 2);
        }
        else 
		sh2.PC += 2;
}

void sh2_opcode_bra()
{
	unsigned long temp, d = current_instruction.d12.disp;
	long disp;
	
	if((d & 0x800) == 0) 
		disp = (0x00000FFF & d);
	else 
		disp = (0xFFFFF000 | d);
	temp = sh2.PC;
	sh2.PC = sh2.PC + (disp << 1) + 4;

	delay_slot(temp+2);

}

void sh2_opcode_braf()
{
	unsigned long temp;

        temp = sh2.PC;
        sh2.PC  += sh2.R[current_instruction.m.source];

        delay_slot(temp + 2);

}

void sh2_opcode_bsr()
{
        long disp;
	long d = current_instruction.d12.disp;

        if((d & 0x800) == 0) 
		disp = (0x00000FFF & d);
        else 
		disp = (0xFFFFF000 | d);
        sh2.PR = sh2.PC;
        sh2.PC = sh2.PC + (disp << 1) + 4;
        delay_slot(sh2.PR + 2);
}

void sh2_opcode_bsrf()
{
	sh2.PR  = sh2.PC;
	sh2.PC += sh2.R[current_instruction.m.source];
	delay_slot(sh2.PR + 2);
}

void sh2_opcode_jmp()
{
	unsigned long temp;
	
	temp = sh2.PC;
	sh2.PC   = sh2.R[current_instruction.m.source] + 4;
	delay_slot(temp + 2);
}

void sh2_opcode_jsr()
{
	sh2.PR = sh2.PC;
	sh2.PC = sh2.R[current_instruction.m.source] + 4;
	delay_slot(sh2.PR + 2);
}

void sh2_opcode_ldcsr()
{
	sh2.SR  = sh2.R[current_instruction.m.source] & 0x000003F3;
	sh2.PC += 2;
}

void sh2_opcode_ldcgbr()
{
	sh2.GBR = sh2.R[current_instruction.m.source];
	sh2.PC += 2;
}

void sh2_opcode_ldcvbr()
{
	sh2.VBR = sh2.R[current_instruction.m.source];
	sh2.PC += 2;
}

void sh2_opcode_ldcmsr()
{
	sh2.SR  = sh2_read_long(sh2.R[current_instruction.m.source]) & 0x000003F3;
	sh2.R[current_instruction.m.source] += 4;
	sh2.PC += 2;
}

void sh2_opcode_ldcmgbr()
{
	sh2.GBR = sh2_read_long(sh2.R[current_instruction.m.source]);
	sh2.R[current_instruction.m.source] += 4;
	sh2.PC += 2;
}

void sh2_opcode_ldcmvbr()
{
	sh2.VBR = sh2_read_long(sh2.R[current_instruction.m.source]);
	sh2.R[current_instruction.m.source] += 4;
	sh2.PC += 2;
}

	
void sh2_opcode_ldsmach()
{
	sh2.MACH = sh2.R[current_instruction.m.source];
	/*
	 		-- This is part of the SH1 Cpu
		if((sh2.MACH & 0x00000200) == 0) 
			sh2.MACH &= 0x000003FF;            
		else 
			sh2.MACH |= 0xFFFFFC00;
	*/
	sh2.PC += 2;
}

void sh2_opcode_ldsmacl()
{
	sh2.MACL = sh2.R[current_instruction.m.source];
	sh2.PC  += 2;
}

void sh2_opcode_ldspr()
{
	sh2.PR = sh2.R[current_instruction.m.source];
	sh2.PC += 2;
}

void sh2_opcode_ldsmmach()
{
	sh2.MACH = sh2_read_long(sh2.R[current_instruction.m.source]);
	sh2.R[current_instruction.m.source] += 4;
	sh2.PC += 2;
}

void sh2_opcode_ldsmmacl()
{
	sh2.MACL = sh2_read_long(sh2.R[current_instruction.m.source]);
	sh2.R[current_instruction.m.source] += 4;
	sh2.PC += 2;
}

void sh2_opcode_ldsmpr()
{
	sh2.PR = sh2_read_long(sh2.R[current_instruction.m.source]);
	sh2.R[current_instruction.m.source] += 4;
	sh2.PC += 2;
}

void sh2_opcode_macl()
{
	long m = current_instruction.nm.source;
	long n = current_instruction.nm.dest;
	unsigned long rnl, rnh, rml, rmh, res0, res1, res2;
	unsigned long temp0, temp1, temp2, temp3;
	long tempm, tempn, fnlml;

	tempn = (long)sh2_read_long(sh2.R[n]);
	sh2.R[n] += 4;
	tempm = (long)sh2_read_long(sh2.R[m]);
	sh2.R[m] += 4;

	if((long)(tempn ^ tempm) < 0)
		fnlml = -1;
	else 
		fnlml = 0;

	if(tempn < 0) 
		tempn = 0 - tempn;
        if(tempm < 0) 
		tempm = 0 - tempm;

	temp1 = (unsigned long)tempn;
	temp2 = (unsigned long)tempm;

	rnl   = temp1 & 0x0000FFFF;
	rnh   = (temp1 >> 16) & 0x0000FFFF;
	rml   = temp2 & 0x0000FFFF;
	rmh   = (temp2 >> 16) & 0x0000FFFF;

	temp0 = rml*rnl;
	temp1 = rmh*rnl;
	temp2 = rml*rnh;
	temp3 = rmh*rnh;

	res2  = 0;
	res1  = temp1 + temp2;
	if(res1 < temp1) 
		res2 += 0x00010000;

	temp1 = (res1 << 16) & 0xFFFF0000;
	res0  = temp0 + temp1;
	if(res0 < temp0) 
		res2++;

	res2  = res2 + ((res1 >> 16) & 0x0000FFFF) + temp3;

	if(fnlml < 0) {
		res2 = ~res2;
		if(res0 == 0) 
			res2++;
		else 
			res0 = (~res0) + 1;
	}
	if(S == 1) {
		res0 = sh2.MACL + res0;
		if(sh2.MACL > res0) 
			res2++;
		res2 += (sh2.MACH & 0x0000FFFF);

		if(((long)res2 < 0) && (res2 < 0xFFFF8000)) {
			res2 = 0x00008000;
			res0 = 0x00000000;
		}
		if(((long)res2 > 0) && (res2 > 0x00007FFF)) {
			res2 = 0x00007FFF;
			res0 = 0xFFFFFFFF;
		};
		sh2.MACH = res2;
		sh2.MACL = res0;
	} else {
		res0 = sh2.MACL + res0;
		if(sh2.MACL > res0) 
			res2++;

		res2 += sh2.MACH;
		sh2.MACH = res2;
		sh2.MACL = res0;
	}
	sh2.PC += 2;
}
 
void sh2_opcode_macw()
{
	long tempm, tempn, dest, src, ans;
	unsigned long templ;
	long n = current_instruction.nm.dest;
	long m = current_instruction.nm.source;

	tempn = (long)sh2_read_word(sh2.R[n]);
	sh2.R[n] += 2;
	tempm = (long)sh2_read_word(sh2.R[m]);
	sh2.R[m] += 2;
	templ = sh2.MACL;
	tempm = ((long)(short)tempn * (long)(short)tempm);

	if((long)sh2.MACL >= 0) 
		dest = 0;
	else 
		dest = 1;
	if((long)tempm >= 0) {
		src = 0;
		tempn = 0;
	} else {
		src = 1;
		tempn = 0xFFFFFFFF;
	}

	src  += dest;
	sh2.MACL += tempm;
	if((long)sh2.MACL >= 0) 
		ans = 0;
	else 
		ans = 1;
	ans  += dest;

	if(S == 1) {
		if(ans == 1) {
			if(src == 0) 
				sh2.MACL = 0x7FFFFFFF;
			if(src == 2) 
				sh2.MACL = 0x80000000;
		}
	} else {
		sh2.MACH += tempn;
		if(templ > sh2.MACL) 
			sh2.MACH += 1;
	}
	sh2.PC += 2;
}

void sh2_opcode_movbs0()
{
	sh2_write_byte(sh2.R[current_instruction.nm.dest] + sh2.R[0],
		       sh2.R[current_instruction.nm.source]);
	sh2.PC += 2;
}

void sh2_opcode_movws0()
{
	sh2_write_word(sh2.R[current_instruction.nm.dest] + sh2.R[0],
		       sh2.R[current_instruction.nm.source]);
	sh2.PC += 2;
}
	                                          
void sh2_opcode_movls0()
{
        sh2_write_long(sh2.R[current_instruction.nm.dest] + sh2.R[0],
                       sh2.R[current_instruction.nm.source]);
        sh2.PC += 2;
}	
         
void sh2_opcode_mov()
{
	sh2.R[current_instruction.nm.dest] = sh2.R[current_instruction.nm.source];
	sh2.PC += 2;
}

void sh2_opcode_movbs()
{
	sh2_write_byte(sh2.R[current_instruction.nm.dest], sh2.R[current_instruction.nm.source]);
	sh2.PC += 2;
}

void sh2_opcode_movws()
{
	sh2_write_word(sh2.R[current_instruction.nm.dest], sh2.R[current_instruction.nm.source]);
	sh2.PC += 2;
}

void sh2_opcode_movls()
{
	sh2_write_long(sh2.R[current_instruction.nm.dest], sh2.R[current_instruction.nm.source]);
	sh2.PC += 2;
}

void sh2_opcode_movbl()
{
	sh2.R[current_instruction.nm.dest] = (long)sh2_read_byte(sh2.R[current_instruction.nm.source]);
	if((sh2.R[current_instruction.nm.dest] & 0x80) == 0) 
		sh2.R[current_instruction.nm.dest] &= 0x000000FF;
	else
		sh2.R[current_instruction.nm.dest] |= 0xFFFFFF00;
	sh2.PC += 2;
}

void sh2_opcode_movwl()
{
	sh2.R[current_instruction.nm.dest] = (long)sh2_read_word(sh2.R[current_instruction.nm.source]);
	if((sh2.R[current_instruction.nm.dest] & 0x8000) == 0)
		sh2.R[current_instruction.nm.dest] &= 0x0000FFFF;
	else
		sh2.R[current_instruction.nm.dest] |= 0xFFFF0000;
	sh2.PC += 2;
}

void sh2_opcode_movll()
{
	sh2.R[current_instruction.nm.dest] = sh2_read_long(sh2.R[current_instruction.nm.source]);
	sh2.PC += 2;
}

void sh2_opcode_movbm()
{
	sh2_write_byte(sh2.R[current_instruction.nm.dest] - 1, 
		       sh2.R[current_instruction.nm.source]);
	sh2.R[current_instruction.nm.dest] -= 1;
	sh2.PC += 2;
}

void sh2_opcode_movwm()
{
	sh2_write_word(sh2.R[current_instruction.nm.dest] - 2,
		       sh2.R[current_instruction.nm.source]);
	sh2.R[current_instruction.nm.dest] -= 2;
	sh2.PC += 2;
}

void sh2_opcode_movlm()
{
	sh2_write_long(sh2.R[current_instruction.nm.dest] - 4,
		       sh2.R[current_instruction.nm.source]);
	sh2.R[current_instruction.nm.dest] -= 4;
	sh2.PC += 2;
}

void sh2_opcode_movi()
{
	long i = current_instruction.ni.imm;
	long n = current_instruction.ni.dest;
	
	if((i & 0x80) == 0) 
		sh2.R[n] = (0x000000FF & (long)i);
	else 
		sh2.R[n] = (0xFFFFFF00 | (long)i);
	sh2.PC += 2;
}

void sh2_opcode_movwi()
{
	long disp;
	long d = current_instruction.nd8.disp;
	long n = current_instruction.nd8.dest;
	
	disp = (0x000000FF & (long)d);
	sh2.R[n] = (long)sh2_read_word(sh2.PC + (disp << 1));
	if((sh2.R[n] & 0x8000) == 0)
		sh2.R[n] &= 0x0000FFFF;
	else 
		sh2.R[n] |= 0xFFFF0000;
	sh2.PC += 2;
}	                              

void sh2_opcode_movli()
{
	long disp;
	long d = current_instruction.nd8.disp;
	long n = current_instruction.nd8.dest;

	disp = (0x000000FF & (long)d);
	sh2.R[n] = sh2_read_long((sh2.PC & 0xFFFFFFFC) + (disp << 2));
	sh2.PC += 2;
}

void sh2_opcode_movbp()
{
	long n = current_instruction.nm.dest;
	long m = current_instruction.nm.source;
	sh2.R[n] = (long)sh2_read_byte(sh2.R[m]);
	if((sh2.R[n] & 0x80) == 0) 
		sh2.R[n] &= 0x000000FF;
	else 
		sh2.R[n] |= 0xFFFFFF00;
	if(n != m)
		sh2.R[m] += 1;
	sh2.PC += 2;
}

void sh2_opcode_movwp()
{
	long n = current_instruction.nm.dest;
	long m = current_instruction.nm.source;
                
	sh2.R[n] = (long)sh2_read_word(sh2.R[m]);
	if((sh2.R[n] & 0x8000) == 0) 
		sh2.R[n] &= 0x0000FFFF;
	else 
		sh2.R[n] |= 0xFFFF0000;
	if(n != m) 
		sh2.R[m] += 2;
	sh2.PC+=2;
}

void sh2_opcode_movlp()
{
	long n = current_instruction.nm.dest;
	long m = current_instruction.nm.source;
               
	sh2.R[n] = sh2_read_long(sh2.R[m]);
	if(n != m)
		sh2.R[m] += 4;
	sh2.PC += 2;
}

void sh2_opcode_movbl0()
{
	long n = current_instruction.nm.dest;
	long m = current_instruction.nm.source;

	sh2.R[n] = (long)sh2_read_byte(sh2.R[m] + sh2.R[0]);
	if((sh2.R[n] & 0x80) == 0) 
		sh2.R[n] &= 0x000000FF;
	else 
		sh2.R[n] |= 0xFFFFFF00;
	sh2.PC += 2;
}

void sh2_opcode_movwl0()
{
	long n = current_instruction.nm.dest;
	long m = current_instruction.nm.source;

	sh2.R[n] = (long)sh2_read_word(sh2.R[m] + sh2.R[0]);
	if((sh2.R[n] & 0x8000) == 0) 
		sh2.R[n] &= 0x0000FFFF;
	else 
		sh2.R[n] |= 0xFFFF0000;
	sh2.PC += 2;
}

void sh2_opcode_movll0()
{
	long n = current_instruction.nm.dest;
	long m = current_instruction.nm.source;
                
	sh2.R[n] = sh2_read_long(sh2.R[m] + sh2.R[0]);
	sh2.PC += 2;
}

void sh2_opcode_movblg()
{
	long disp;
	long d = current_instruction.d.disp;
      
	disp = (0x000000FF & (long)d);
	sh2.R[0] = (long)sh2_read_byte(sh2.GBR + disp);
	if((sh2.R[0] & 0x80) == 0) 
		sh2.R[0] &= 0x000000FF;
	else 
		sh2.R[0] |= 0xFFFFFF00;
	sh2.PC += 2;
}

void sh2_opcode_movwlg()
{
	long disp;
	long d = current_instruction.d.disp;      
      
	disp = (0x000000FF & (long)d);
	sh2.R[0] = (long)sh2_read_word(sh2.GBR + (disp << 1));
	if((sh2.R[0] & 0x8000) == 0) 
		sh2.R[0] &= 0x0000FFFF;
	else 
		sh2.R[0] |= 0xFFFF0000;
	sh2.PC += 2;
}

void sh2_opcode_movllg()
{
	long disp;
	long d = current_instruction.d.disp;

	disp = (0x000000FF & (long)d);
	sh2.R[0] = sh2_read_long(sh2.GBR + (disp << 2));
	sh2.PC += 2;
}
                  	                                                                        
void sh2_opcode_movbsg()
{
	long disp;
	long d = current_instruction.d.disp;
	
	disp = (0x000000FF & (long)d);
	sh2_write_byte(sh2.GBR + disp, sh2.R[0]);
	sh2.PC += 2;
}
	                  
void sh2_opcode_movwsg()
{
	long disp;
	long d = current_instruction.d.disp;
	
	disp = (0x000000FF & (long)d);
	sh2_write_word(sh2.GBR + (disp << 1), sh2.R[0]);
	sh2.PC += 2;
}

void sh2_opcode_movlsg()
{
	long disp;
	long d = current_instruction.d.disp;
	
	disp = (0x000000FF & (long)d);
	sh2_write_long(sh2.GBR+(disp<<2),sh2.R[0]);
	sh2.PC += 2;
}
                  	
void sh2_opcode_movbs4()
{
	long disp;
	long d = current_instruction.nd4.disp;
	long n = current_instruction.nd4.dest;
      
	disp = (0x0000000F & (long)d);
	sh2_write_byte(sh2.R[n] + disp, sh2.R[0]);
	sh2.PC += 2;
}
                        	
void sh2_opcode_movws4()
{
	long disp;
	long d = current_instruction.nd4.disp;
	long n = current_instruction.nd4.dest;

	disp = (0x0000000F & (long)d);
	sh2_write_word(sh2.R[n] + (disp << 1), sh2.R[0]);
	sh2.PC += 2;
}
                                          
void sh2_opcode_movls4()
{
	long disp;
	long d = current_instruction.nmd.disp;
	long n = current_instruction.nmd.dest;
	long m = current_instruction.nmd.source;
	
	disp = (0x0000000F & (long)d);
	sh2_write_long(sh2.R[n] + (disp << 2), sh2.R[m]);
	sh2.PC += 2;
}

void sh2_opcode_movbl4()
{
	long disp;
	long d = current_instruction.md.disp;
	long m = current_instruction.md.source;

	disp = (0x0000000F & (long)d);
	sh2.R[0] = sh2_read_byte(sh2.R[m] + disp);
	if((sh2.R[0] & 0x80) == 0) 
		sh2.R[0] &= 0x000000FF;
	else 
		sh2.R[0] |= 0xFFFFFF00;
	sh2.PC += 2;
}

void sh2_opcode_movwl4()
{
	long disp;
	long d = current_instruction.md.disp;
	long m = current_instruction.md.source;
	
	disp = (0x0000000F & (long)d);
	sh2.R[0] = sh2_read_word(sh2.R[m] + (disp << 1));
	if((sh2.R[0] & 0x8000) == 0) 
		sh2.R[0] &= 0x0000FFFF;
	else 
		sh2.R[0] |= 0xFFFF0000;
	sh2.PC += 2;
}

void sh2_opcode_movll4()
{
        long disp;
        long d = current_instruction.nmd.disp;
        long n = current_instruction.nmd.dest;
        long m = current_instruction.nmd.source;

	disp = (0x0000000F & (long)d);
	sh2.R[n] = sh2_read_long(sh2.R[m] + (disp << 2));
	sh2.PC += 2;
}

void sh2_opcode_mova()
{
	long disp;
	
	disp = (0x000000FF & (long)current_instruction.d.disp);
	sh2.R[0] = (sh2.PC & 0xFFFFFFFC) + (disp << 2);
	sh2.PC += 2;
}

void sh2_opcode_movt()
{
	sh2.R[current_instruction.n.dest] = (0x00000001 & sh2.SR);
	sh2.PC += 2;
}

void sh2_opcode_mull()
{
	sh2.MACL = sh2.R[current_instruction.nm.dest] * sh2.R[current_instruction.nm.source];
	sh2.PC += 2;
}

void sh2_opcode_muls()
{
	sh2.MACL = ((long)(short)sh2.R[current_instruction.nm.dest] * (long)(short)sh2.R[current_instruction.nm.source]);
	sh2.PC += 2;
}

void sh2_opcode_mulu()
{
	sh2.MACL = ((unsigned long)(unsigned short)sh2.R[current_instruction.nm.dest]
		* (unsigned long)(unsigned short)sh2.R[current_instruction.nm.source]);
	sh2.PC += 2;
}

void sh2_opcode_neg()
{
	sh2.R[current_instruction.nm.dest] = 0 - sh2.R[current_instruction.nm.source];
	sh2.PC += 2;
}

void sh2_opcode_negc()
{
	unsigned long temp;
	
	temp = 0 - sh2.R[current_instruction.nm.source];
	sh2.R[current_instruction.nm.dest] = temp - T;
	if(0 < temp)
		T = 1;
	else
		T = 0;
	if(temp < sh2.R[current_instruction.nm.dest])
		T = 1;
	sh2.PC += 2;
}

void sh2_opcode_not()
{
	sh2.R[current_instruction.nm.dest] = ~sh2.R[current_instruction.nm.source];
	sh2.PC += 2;
}

void sh2_opcode_or()
{
	sh2.R[current_instruction.nm.dest] |= sh2.R[current_instruction.nm.source];
	sh2.PC += 2;
}

void sh2_opcode_ori()
{
	sh2.R[0] |= (0x000000FF & (long)current_instruction.i.imm);
	sh2.PC += 2;
}

void sh2_opcode_orm()
{
	long temp;
	
	temp  = (long)sh2_read_byte(sh2.GBR + sh2.R[0]);
	temp |= (0x000000FF & (long)current_instruction.i.imm);
	sh2_write_byte(sh2.GBR + sh2.R[0], temp);
	sh2.PC += 2;
}

void sh2_opcode_rotcl()
{
	long temp;
	long n = current_instruction.n.dest;
	
	if((sh2.R[n] & 0x80000000) == 0) 
		temp = 0;
	else 
		temp = 1;

	sh2.R[n] <<= 1;

	if(T == 1) 
		sh2.R[n] |= 0x00000001;
	else 
		sh2.R[n] &= 0xFFFFFFFE;

	if(temp == 1) 
		T = 1;
	else 
		T = 0;
	sh2.PC += 2;
}

void sh2_opcode_rotcr()
{
	long temp;
	long n = current_instruction.n.dest;
	
	if((sh2.R[n] & 0x00000001) == 0) 
		temp = 0;
	else 
		temp = 1;

	sh2.R[n] >>= 1;
	if(T == 1) 
		sh2.R[n] |= 0x80000000;
	else 
		sh2.R[n] &= 0x7FFFFFFF;
	if(temp == 1) 
		T = 1;
	else 
		T = 0;
	sh2.PC += 2;
}

void sh2_opcode_rotl()
{
	long n = current_instruction.n.dest;
	if((sh2.R[n] & 0x80000000) == 0) 
		T = 0;
	else 
		T = 1;

	sh2.R[n] <<= 1;
	if(T == 1) 
		sh2.R[n] |= 0x00000001;
	else 
		sh2.R[n] &= 0xFFFFFFFE;

	sh2.PC += 2;
}

void sh2_opcode_rotr()
{
	long n = current_instruction.n.dest;
	if((sh2.R[n] & 0x00000001) == 0) 
		T = 0;
	else
		T = 1;

	sh2.R[n] >>= 1;

	if(T == 1) 
		sh2.R[n] |= 0x80000000;
	else 
		sh2.R[n] &= 0x7FFFFFFF;

	sh2.PC += 2;
}

void sh2_opcode_rte()
{
	unsigned long temp;
	temp = sh2.PC;
	
	sh2.PC = sh2_read_long(sh2.R[15]) + 4;
	sh2.R[15] += 4;
	
	sh2.SR = sh2_read_long(sh2.R[15]) & 0x000003F3;
	sh2.R[15] += 4;
	
	delay_slot(temp + 2);
}

void sh2_opcode_rts()
{
	unsigned long temp;
	
	temp = sh2.PC;
	sh2.PC = sh2.PR + 4;

	delay_slot(temp + 2);
}

void sh2_opcode_shal()
{
	if((sh2.R[current_instruction.n.dest] & 0x80000000) == 0)
		T = 0;
	else
		T = 1;
		
	sh2.R[current_instruction.n.dest] <<= 1;
	sh2.PC += 2;
}

void sh2_opcode_shar()
{
	long temp;
	
	if((sh2.R[current_instruction.n.dest] & 0x00000001) == 0)
		T = 0;
	else
		T = 1;
	
	if((sh2.R[current_instruction.n.dest] & 0x80000000) == 0)
		temp = 0;
	else
		temp = 1;
	
	sh2.R[current_instruction.n.dest] >>= 1;
	
	if(temp == 1)
		sh2.R[current_instruction.n.dest] |= 0x80000000;
	else
		sh2.R[current_instruction.n.dest] &= 0x7FFFFFFF;
		
	sh2.PC += 2;
}

void sh2_opcode_shll2()
{
	sh2.R[current_instruction.n.dest] <<= 2;
	sh2.PC += 2;
}

void sh2_opcode_shll8()
{
	sh2.R[current_instruction.n.dest] <<= 8;
	sh2.PC += 2;
}

void sh2_opcode_shll16()
{
	sh2.R[current_instruction.n.dest] <<= 16;
	sh2.PC += 2;
}

void sh2_opcode_shlr()
{
	if((sh2.R[current_instruction.n.dest] & 0x00000001) == 0)
		T = 0;
	else
		T = 1;
	sh2.R[current_instruction.n.dest] >>= 1;
	sh2.R[current_instruction.n.dest]  &= 0x7FFFFFFF;
	sh2.PC += 2;
}

void sh2_opcode_shlr2()
{
	sh2.R[current_instruction.n.dest] >>= 2;
	sh2.R[current_instruction.n.dest]  &= 0x3FFFFFFF;
	sh2.PC += 2;
}

void sh2_opcode_shlr8()
{
	sh2.R[current_instruction.n.dest] >>= 8;
	sh2.R[current_instruction.n.dest]  &= 0x00FFFFFF;
	sh2.PC += 2;
}

void sh2_opcode_shlr16()
{
	sh2.R[current_instruction.n.dest] >>= 16;
	sh2.R[current_instruction.n.dest]  &= 0x0000FFFF;
	sh2.PC += 2;
}

void sh2_opcode_stcsr()
{
	sh2.R[current_instruction.n.dest] = sh2.SR;
	sh2.PC += 2;
}

void sh2_opcode_stcgbr()
{
	sh2.R[current_instruction.n.dest] = sh2.GBR;
	sh2.PC += 2;
}

void sh2_opcode_stcvbr()
{
	sh2.R[current_instruction.n.dest] = sh2.VBR;
	sh2.PC += 2;
}

void sh2_opcode_stcmsr()
{
	sh2.R[current_instruction.n.dest] -= 4;
	sh2_write_long(sh2.R[current_instruction.n.dest], sh2.SR);
	sh2.PC += 2;
}

void sh2_opcode_stcmgbr()
{
	sh2.R[current_instruction.n.dest] -= 4;
	sh2_write_long(sh2.R[current_instruction.n.dest], sh2.GBR);
	sh2.PC += 2;
}

void sh2_opcode_stcmvbr()
{
	sh2.R[current_instruction.n.dest] -= 4;
	sh2_write_long(sh2.R[current_instruction.n.dest], sh2.VBR);
	sh2.PC += 2;
}

void sh2_opcode_stsmach()
{
	sh2.R[current_instruction.n.dest] = sh2.MACH;

	/* SH-1 here *\
	 if((sh1.R[current_instruction.n.dest] & 0x00000200) == 0)
	 	sh1.R[current_instruction.n.dest] &= 0x000003FF;
	 else
	 	sh1.R[current_instruction.n.dest] |= 0xFFFFFC00;
	\* --------- */

	sh2.PC += 2;
}
	
void sh2_opcode_stsmacl()
{
	sh2.R[current_instruction.n.dest] = sh2.MACL;
	sh2.PC += 2;
}

void sh2_opcode_stspr()
{
	sh2.R[current_instruction.n.dest] = sh2.PR;
	sh2.PC += 2;
}

void sh2_opcode_stsmmach()
{
	sh2.R[current_instruction.n.dest] -= 4;

	/* for sh-1 *\
	  if((sh1.MACH & 0x00000200) == 0)
	  	sh1_write_long(sh1.R[current_instruction.n.dest], sh1.MACH & 0x000003FF);
	  else
	  	sh1_write_long(sh1.R[current_instruction.n.dest], sh1.MACH | 0xFFFFFC00);
	\* -------- */
	
	sh2_write_long(sh2.R[current_instruction.n.dest], sh2.MACH);
	sh2.PC += 2;
}

void sh2_opcode_stsmmacl()
{
	sh2.R[current_instruction.n.dest] -= 4;
	sh2_write_long(sh2.R[current_instruction.n.dest], sh2.MACL);
	sh2.PC += 2;
}

void sh2_opcode_stsmpr()
{
	sh2.R[current_instruction.n.dest] -= 4;
	sh2_write_long(sh2.R[current_instruction.n.dest], sh2.PR);
	sh2.PC += 2;
}

void sh2_opcode_sub()
{
	sh2.R[current_instruction.nm.dest] -= sh2.R[current_instruction.nm.source];
	sh2.PC += 2;
}

void sh2_opcode_subc()
{
	unsigned long tmp0, tmp1;
	
	tmp1 = sh2.R[current_instruction.nm.dest] - sh2.R[current_instruction.nm.source];
	tmp0 = sh2.R[current_instruction.nm.dest];
	sh2.R[current_instruction.nm.dest] = tmp1 - T;
	
	if(tmp0 < tmp1)
		T = 1;
	else
		T = 0;
	
	if(tmp1 < sh2.R[current_instruction.nm.dest]) 
		T = 1;
	
	sh2.PC += 2;
}

void sh2_opcode_subv()
{
	long n = current_instruction.nm.dest;
	long m = current_instruction.nm.source;
	long dest, src, ans;
	
	if((long)sh2.R[n] >= 0)
		dest = 0;
	else
		dest = 1;
	if((long)sh2.R[m] >= 0)
		src = 0;
	else
		src = 1;
	
	src += dest;
	
	sh2.R[n] -= sh2.R[m];
	
	if((long)sh2.R[n] >= 0)
		ans = 0;
	else
		ans = 1;
	
	ans += dest;
	
	if(src == 1) {
		if(ans == 1)
			T = 1;
		else
			T = 0;
	} else 
		T = 0;
	sh2.PC += 2;
}

void sh2_opcode_swapb()
{
	long m = current_instruction.nm.source;
	long n = current_instruction.nm.dest;
	
	sh2.R[n] = (sh2.R[m] & 0xFFFF0000) | 
		   ((sh2.R[m] & 0x000000FF) << 8) |
	           ((sh2.R[m] & 0x0000FF00) >> 8);

	sh2.PC += 2;
}

void sh2_opcode_swapw()
{
	long m = current_instruction.nm.source;
	long n = current_instruction.nm.dest;

	sh2.R[n] = ((sh2.R[m] & 0xFFFF0000) >> 16) |
		   ((sh2.R[m] & 0x0000FFFF) << 16);

	sh2.PC += 2;
}

void sh2_opcode_tas()
{
	long temp;
	
	temp = (long) sh2_read_byte(sh2.R[current_instruction.n.dest]);
	if(temp == 0)
		T = 1;
	else
		T = 0;
	
	temp |= 0x00000080;
	
	sh2_write_byte(sh2.R[current_instruction.n.dest], temp);
	sh2.PC += 2;
}

void sh2_opcode_trapa()
{
	long imm;
	long i = current_instruction.i.imm;
	
	imm = (0x000000FF & i);
	sh2.R[15] -= 4;
	sh2_write_long(sh2.R[15], sh2.SR);
	
	sh2.R[15] -= 4;
	sh2_write_long(sh2.R[15], sh2.PC - 2);
	sh2.PC = sh2_read_long(sh2.VBR + (imm << 2)) + 4;
}

void sh2_opcode_tst()
{

	if((sh2.R[current_instruction.nm.dest] & 
	    sh2.R[current_instruction.nm.source]) == 0) 
	    	T = 1;
	else 
		T = 0;

	sh2.PC += 2;
}

void sh2_opcode_tsti()
{
	long temp;

	temp = sh2.R[0] & (0x000000FF & (long)current_instruction.i.imm);
	if(temp == 0)
		T = 1;
	else
		T = 0;

	sh2.PC += 2;
}

void sh2_opcode_tstm()
{
	long temp;
	
	temp = (long)sh2_read_byte(sh2.GBR + sh2.R[0]);
	temp &= (0x000000FF & (long)current_instruction.i.imm);

	if(temp == 0)
		T = 1;
	else
		T = 0;
	
	sh2.PC += 2;
}

void sh2_opcode_xor()
{
	sh2.R[current_instruction.nm.dest] ^=
		sh2.R[current_instruction.nm.source];
	sh2.PC += 2;
}

void sh2_opcode_xori()
{
	sh2.R[0] ^= (0x000000FF & (long)current_instruction.i.imm);
	sh2.PC += 2;
}

void sh2_opcode_xorm()
{
	long temp;
	
	temp  = (long)sh2_read_byte(sh2.GBR + sh2.R[0]);
	temp ^= (0x000000FF & (long)current_instruction.i.imm);
	
	sh2_write_byte(sh2.GBR + sh2.R[0], temp);
	sh2.PC += 2;
}

void sh2_opcode_xtrct()
{
	unsigned long temp;
	
	temp = (sh2.R[current_instruction.nm.source] << 16) & 0xFFFF0000;
	sh2.R[current_instruction.nm.dest] = (sh2.R[current_instruction.nm.dest] >> 16) & 0x0000FFFF;
	sh2.R[current_instruction.nm.dest] |= temp;
	sh2.PC += 2;
}
