/*
    semu - Sega Saturn Emulator
    Copyright (C) 1998  Chuck Mason

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Chuck Mason IV                      chuckjr@sundial.net
    477 Moffat Loop
    Oviedo, Fl 32765
*/

#ifndef __OPCODES_H
#define __OPCODES_H

#define TOTAL_OPCODE_COUNT 142
#define OPCODE_COUNT 142

typedef struct _instruction_list {
        unsigned short int mask;
        unsigned short int opcode;
        int format;
        void (*execute)();
        char *instrformat;
} instruction_list;

extern instruction_list sh2_instruction_list[];
void sh2_format_instruction_string(unsigned long, char *);

/* This will go in dyninstr.c */
extern instruction_list sh2_instructions[];
void sh2_build_instruction_list();
/* ************************** */

void sh2_opcode_nop();
void sh2_opcode_add();
void sh2_opcode_addi();
void sh2_opcode_addc();
void sh2_opcode_addv();
void sh2_opcode_and();
void sh2_opcode_andi();
void sh2_opcode_andm();
void sh2_opcode_bf();
void sh2_opcode_bt();
void sh2_opcode_clrmac();

void sh2_opcode_clrt();
void sh2_opcode_sett();

void sh2_opcode_cmpeq();
void sh2_opcode_cmpge();
void sh2_opcode_cmpgt();
void sh2_opcode_cmphi();
void sh2_opcode_cmphs();
void sh2_opcode_cmppl();
void sh2_opcode_cmppz();
void sh2_opcode_cmpstr();
void sh2_opcode_cmpim();

void sh2_opcode_div0s();
void sh2_opcode_div0u();
void sh2_opcode_div1();

void sh2_opcode_dmuls();
void sh2_opcode_dmulu();

void sh2_opcode_dt();

void sh2_opcode_extsb();
void sh2_opcode_extsw();
void sh2_opcode_extub();
void sh2_opcode_extuw();

void sh2_opcode_sleep();
void sh2_opcode_invalid();

void sh2_opcode_bfs();
void sh2_opcode_bts();

void sh2_opcode_bra();
void sh2_opcode_braf();
void sh2_opcode_bsr();
void sh2_opcode_bsrf();

void sh2_opcode_jmp();
void sh2_opcode_jsr();

void sh2_opcode_ldcsr();
void sh2_opcode_ldcgbr();
void sh2_opcode_ldcvbr();
void sh2_opcode_ldcmsr();
void sh2_opcode_ldcmgbr();
void sh2_opcode_ldcmvbr();

void sh2_opcode_ldsmach();
void sh2_opcode_ldsmacl();
void sh2_opcode_ldspr();
void sh2_opcode_ldsmmach();
void sh2_opcode_ldsmmacl();
void sh2_opcode_ldsmpr();

void sh2_opcode_macl();
void sh2_opcode_macw();

/* MOV Series!!#@$!@#$ */
void sh2_opcode_mov();
void sh2_opcode_movbs();
void sh2_opcode_movws();
void sh2_opcode_movls();
void sh2_opcode_movbl();
void sh2_opcode_movwl();
void sh2_opcode_movll();
void sh2_opcode_movbm();
void sh2_opcode_movwm();
void sh2_opcode_movlm();
void sh2_opcode_movi();
void sh2_opcode_movwi();
void sh2_opcode_movli();
void sh2_opcode_movbp();
void sh2_opcode_movwp();
void sh2_opcode_movlp();
void sh2_opcode_movbs0();
void sh2_opcode_movws0();
void sh2_opcode_movls0();
void sh2_opcode_movbl0();
void sh2_opcode_movwl0();
void sh2_opcode_movll0();

void sh2_opcode_movblg();
void sh2_opcode_movwlg();
void sh2_opcode_movllg();
void sh2_opcode_movbsg();
void sh2_opcode_movwsg();
void sh2_opcode_movlsg();

void sh2_opcode_movbs4();
void sh2_opcode_movws4();
void sh2_opcode_movls4();
void sh2_opcode_movbl4();
void sh2_opcode_movwl4();
void sh2_opcode_movll4();

void sh2_opcode_mova();

void sh2_opcode_movt();
/* ------------------------ That's it for the MOV series!! */
void sh2_opcode_mull();
void sh2_opcode_muls();
void sh2_opcode_mulu();

void sh2_opcode_neg();
void sh2_opcode_negc();

void sh2_opcode_not();

void sh2_opcode_or();
void sh2_opcode_ori();
void sh2_opcode_orm();

void sh2_opcode_rotcl();
void sh2_opcode_rotcr();

void sh2_opcode_rotl();
void sh2_opcode_rotr();

void sh2_opcode_rte();
void sh2_opcode_rts();

void sh2_opcode_shal();
void sh2_opcode_shar();

void sh2_opcode_shll2();
void sh2_opcode_shll8();
void sh2_opcode_shll16();

void sh2_opcode_shlr();

void sh2_opcode_shlr2();
void sh2_opcode_shlr8();
void sh2_opcode_shlr16();

void sh2_opcode_stcsr();
void sh2_opcode_stcgbr();
void sh2_opcode_stcvbr();
void sh2_opcode_stcmsr();
void sh2_opcode_stcmgbr();
void sh2_opcode_stcmvbr();

void sh2_opcode_stsmach();
void sh2_opcode_stsmacl();
void sh2_opcode_stspr();
void sh2_opcode_stsmmach();
void sh2_opcode_stsmmacl();
void sh2_opcode_stsmpr();

void sh2_opcode_sub();
void sh2_opcode_subc();
void sh2_opcode_subv();

void sh2_opcode_swapb();
void sh2_opcode_swapw();

void sh2_opcode_tas();

void sh2_opcode_trapa();

void sh2_opcode_tst();
void sh2_opcode_tsti();
void sh2_opcode_tstm();

void sh2_opcode_xor();
void sh2_opcode_xori();
void sh2_opcode_xorm();

void sh2_opcode_xtrct();

#endif
